// T3h Package.
package kirk.gui.dialog;

// Import t3h AWT.
import java.awt.*;
import java.awt.event.ActionEvent;

/**
 * An extension of <code>PopupDialog</code> that asks the user for input. This class
 * is abstract, the actual input is asked by subclasses.
 * The method <code>actionPerformed</code> assumes there is an "OK" button, or a
 * "Cancel" button. In case of the "OK" button, it calls the <code>retrieve</code>
 * method in the <code>Retriever</code> given on construction time. This method
 * should then retrieve any data wanted, and return. This class will then close
 * itself if <code>true</code> was returned by <code>retrieve</code>. If the action
 * command specified by <code>ActionEvent.getActionCommand()</code> is "Cancel" (case
 * insensitive), then the <code>PopupDialog</code> will be closed without calling the
 * <code>retrieve</code> method. In all other cases of action command, this class
 * assumes an "OK" button. 
 * <br>
 * The implementation of <code>actionPerformed</code> in this class allows for
 * subclasses that only define a specific look for the input window, and that have
 * methods and/or attributes that can be used/retrieved by the <code>retrieve</code>
 * method. All buttons that they have should act like "OK" or "Cancel" buttons (see
 * above). If all these requirements have been met, the subclass doesn't have to
 * override <code>actionPerformed</code>. If it needs more buttons (say "Clear")
 * you need to override <code>actionPerformed</code>. You can then define your own
 * button usage, and use <code>super</code> to activate the "OK" and "Cancel" buttons
 * through this class. Note that this will make all buttons that aren't caught using
 * action commands will default to "OK".<br>
 * <br>
 * If you close an <code>InputDialog</code> with the close button inherent to a
 * <code>Window</code>, closing is done through a standard <code>WindowCloser</code>.
 * <i>This means that this button has the same effect as a "Cancel" button</i>.
 * This means <code>retrieve</code> will <i>not</i> be called. This behaviour is by
 * design, always providing a means to cancel.
 * 
 * @author Berend "Kirk" Wouda
 * @version 1
 * @see kirk.gui.dialog.Retriever
 */
public abstract class InputDialog extends PopupDialog {
	/**
	 * Constructs a new <code>InputDialog</code> with the passed settings.
	 * 
	 * @param owner The pwning <code>Frame</code>.
	 * @param retriever The <code>Retriever</code> that retrieves the values from
	 * this object.
	 * @param width The width of this <code>Dialog</code> in pixels.
	 * @param height The height of this <code>Dialog</code> in pixels.
	 */
	public InputDialog(Frame owner, Retriever retriever, int width, int height) {
		// Call the super constructor.
		super(owner, "Input Dialog", width, height);
		
		// Set the retriever.
		this.retriever = retriever;
	}
	

	/**
	 * This method gets called when an ActionEvent is generated. By default, this
	 * object assumes this is caused by an OK button. So, it makes the
	 * <code>Retriever</code> retrieve the values it requires, and it then calls the
	 * super method of <code>PopupDialog</code>, which closes this
	 * <code>Dialog</code>.
	 * 
	 * However, if the action command of the button is "Cancel" (which is case
	 * insensitive), the <code>Dialog</code> will be closed as if the close button
	 * was pressed. The <code>Retriever</code> will <i>not</i> be called.
	 *  
	 * Subclasses of this class should supercall this method when requiring this
	 * functionality.
	 * 
	 * @param event The standard passed <code>ActionEvent</code>.
	 * @see kirk.gui.dialog.PopupDialog#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent event) {
		// Retrieve the action command.
		String command = event.getActionCommand();
		
		// Check whether the action command is a negative.
		if(command.equalsIgnoreCase("Cancel")) {
			// Just call PopupDialog to close the window.
			super.actionPerformed(event);
		}
		// In any other case, we default to the behaviour of an affirmative.
		else {
			// Make the retriever retrieve.
			// If true is returned, close this InputDialog.
			if(retriever.retrieve(this))  super.actionPerformed(event);
		}
	}
	
	
	/**
	 * The <code>Retriever</code>.
	 */
	private Retriever retriever;
}