// Do t3h monkah!
package kirk.io.sdl;

// Import the io classes.
import java.io.*;

/**
 * This class is able to read ESDL files into <code>SDLDocument</code>s. To make
 * things as easy as possible, all you have to do is:<br>
 * <br>
 * <code>SDLDocument sdldocument = ESDLReader.readESDL(&lt;InputStream&gt;);</code>
 * (when reading from an <code>InputStream</code>)<br>
 * or<br>
 * <code>SDLDocument sdldocument = ESDLReader.readESDLFile(&lt;filename&gt;);</code>
 * (when reading from a file).<br>
 * <br>
 * This also means you cannot instantiate this class yourself. If for some reason you
 * need to do this, extend this class. The constructor is protected.
 * 
 * Note that the ESDL data must be encoded with UTF-16.
 * 
 * @author Berend "Kirk" Wouda
 * @version 1
 * @see kirk.io.sdl.SDLDocument
 */
public class ESDLReader {
	/**
	 * Constructs a new <code>ESDLReader</code> with the given input source to read
	 * from. If you wish to read from a file for example, supply a
	 * <code>new FileInputStream("filename")</code>. The input source should provide
	 * raw byte data, which is converted to characters using UTF-16 encoding.
	 * 
	 * @param in The <code>InputStream</code> to be read from.
	 * @throws IOException When an IO exception occurs. Mmmyeah.
	 */
	protected ESDLReader(InputStream in) throws IOException {
		// Set the Reader. It is a BufferedReader which takes its data from an
		// InputStreamReader, which uses UTF-16 as charset. So the incoming data must
		// be in UTF-16 format, otherwise it will not be read right.
		// If for some reason the JVM we're running on doesn't recognize UTF-16, an
		// UnsupportedEncodingException will be thrown. This is however an
		// IOException, so we won't specifically throw it. 
		reader = new BufferedReader(new InputStreamReader(in, "UTF-16"));
	}


	/**
	 * Reads the ESDL document from the given source and returns it as an
	 * <code>SDLDocument</code> object.
	 * 
	 * @return An <code>SDLDocument</code> tree with all the ESDL data from the file.
	 * @throws IOException When IO errors occur (duh).
	 * @throws SDLReadException When the ESDL is not correct.
	 */
	protected SDLDocument readESDLDocument() throws IOException, ESDLReadException {
		// Let t3h reading begin!
		// This is how it works:
		// First we read an element. Then we read the body of that element, and then
		// we read the next element. Since an ESDL file has only one root element, we
		// can start this process once and then, after that, the file should be
		// entirely read in.
		// Read the root element (which will make the file recusively be read in).
		SDLElement root = readElement();
		
		// Since this is ESDL, there is stuff allowed after the root element.
		// However, only whitespace and comments :P
		skipWhiteSpaceAndComments();
		
		// Now the stream should have no more characters.
		if(reader.read() == -1) {
			// Close the stream.
			reader.close();
			
			// Return a document with root as root (:P).
			return new SDLDocument(root);
		}
		else {
			// T3h suxx0r. The file stream provided a correct ESDL description, but
			// there's more. That is not allowed.
			throw getException("End of file expected.");
		}
	}
	
	/**
	 * Reads the element at the current position in the file, and the contained
	 * value. Then it returns it as an <code>SDLElement</code>. The current position
	 * in the file has to be followed by &lt;element&gt; or &lt;data&gt;, or those
	 * with whitespace and/or comments before it.
	 * 
	 * @return The element at the current file position.
	 * @throws IOException When IO errors occur (duh).
	 * @throws SDLReadException When the ESDL is not correct.
	 */
	protected SDLElement readElement() throws IOException, ESDLReadException {
		// The element at the current position.
		SDLElement element;
		
		// The type of the current element.
		String type;
		
		// Since this is ESDL, there is whitespace and comments allowed before
		// the next tag.
		skipWhiteSpaceAndComments();
		
		// Check for the element or data tag.
		if(checkStartTag(SDLNormalElement.IDENTIFIER)) {
			// Set the type.
			type = SDLNormalElement.IDENTIFIER;
			
			// Read in a normal element.
			element = readNormalElement();
		} 
		else if(checkStartTag(SDLDataElement.IDENTIFIER)) {
			// Set the type.
			type = SDLDataElement.IDENTIFIER;
			
			// Read in a data element.
			element = readDataElement();
		}
		else {
			// The element tag is not there.
			throw getException("<" + SDLNormalElement.IDENTIFIER + "> or <" + SDLDataElement.IDENTIFIER + "> expected.");
		}
		
		// Since this is ESDL, there is whitespace and comments allowed before
		// the next tag.
		skipWhiteSpaceAndComments();
		
		// Check whether the element is closed correctly.
		if(checkEndTag(type)) {
			// This element is correct. Return it.
			return element;
		}
		else {
			// This element is not correctly closed!
			throw getException("<" + type + "> does not have a closing tag.");
		}
	}
	
	/**
	 * Reads the normal element at the current position in the file, and the
	 * contained value. Then it returns it as an <code>SDLDNormalElement</code>. The
	 * current position in the file has NOT to be followed by &lt;element&gt;. This
	 * has already been made sure in <code>readElement()</code>. There may be
	 * whitespace and/or comment at the current position as well.
	 * 
	 * This method will only read until the end of the value.
	 * <code>readElement()</code> will then check for the end tag.
	 * 
	 * @return The normal element at the current file position.
	 * @throws IOException When IO errors occur (duh).
	 * @throws ESDLReadException When the ESDL is not correct.
	 */
	protected SDLNormalElement readNormalElement() throws IOException, ESDLReadException {
		// <element> has already been read, so the next thing we expect is the name
		// tag pair.
		// But, since this is ESDL, there is whitespace and comments allowed before
		// the next tag.
		skipWhiteSpaceAndComments();
		
		// The name of that next tag.
		String name = readTagValue("name");
		
		// The parent (current) element.
		SDLNormalElement parent = new SDLNormalElement(name);
		
		// Skip the stuff ESDL allows.
		skipWhiteSpaceAndComments();
		
		// After that, we expect the value tag pair.
		if(checkStartTag("value")) {
			// Since this is ESDL, there is whitespace and comments allowed before
			// the next tag.
			skipWhiteSpaceAndComments();
			
			// Now we need to read elements until we encounter the end tag.
			while(!checkEndTag("value")) {
				// Read the next (child) element and add it to the parent element.
				parent.addValue(readElement());
				
				// Since this is ESDL, there is whitespace and comments allowed
				// before the next tag.
				skipWhiteSpaceAndComments();
			}
			
			// The value end tag was apparently read. Return the parent element.
			return parent;
		}
		else {
			// No value opening tag.
			throw getException("Opening tag <value> expected.");
		}
	}
	
	/**
	 * Reads the data element at the current position in the file, and the contained
	 * value. Then it returns it as an <code>SDLDataElement</code>. The current
	 * position in the file has NOT to be followed by &lt;data&gt;. This has already
	 * been made sure in <code>readElement()</code>. There may be whitespace and/or
	 * comment at the current position as well.
	 * 
	 * This method will only read until the end of the value.
	 * <code>readElement()</code> will then check for the end tag.
	 * 
	 * @return The data element at the current file position.
	 * @throws IOException When IO errors occur (duh).
	 * @throws ESDLReadException When the ESDL is not correct.
	 */
	protected SDLDataElement readDataElement() throws IOException, ESDLReadException {
		// <data> has already been read, so the next thing we expect is the name tag
		// pair. After that, we expect the value tag pair.
		// But, since this is ESDL, there is whitespace and comments allowed before
		// the next tag.
		skipWhiteSpaceAndComments();
		
		// So, read those in that order from the description, and use it to construct
		// a data element. And return that, of course.
		return new SDLDataElement(readTagValue("name"), readTagValue("value"));
	}
	
	/**
	 * Return the value inbetween the tag pair (start and end tag) with the given
	 * name. Whitespace and comments may precede the start tag of the given tag pair.
	 * 
	 * @param name The name of the tag.
	 * @return The value of the tag pair.
	 * @throws IOException When IO errors occur (duh).
	 * @throws ESDLReadException When the ESDL is not correct.
	 */
	private String readTagValue(String name) throws IOException, ESDLReadException {
		// Since this is ESDL, there is whitespace and comments allowed before the
		// next tag.
		skipWhiteSpaceAndComments();
		
		// First a name start tag needs to be present.
		if(checkStartTag(name)) {
			// The StringBuffer that contains the value.
			StringBuffer value = new StringBuffer();
			
			// Now we need to read until we encounter the end tag.
			while(!checkEndTag(name)) {
				// Read the next character and put it at the end of the value.
				value.append(read());
			}
			
			// The end tag was apparently read. Return the read value as a String.
			return value.toString();
		}
		else {
			// No opening tag.
			throw getException("Opening tag <" + name + "> expected.");
		}
	}
	
	
	/**
	 * This method is especially for ESDL. It skips all the comment and whitespace
	 * found at the current file position, until it encounters a non-whitespace
	 * character that is not part of comment.
	 * 
	 * @throws IOException When IO errors occur (duh).
	 * @throws ESDLReadException When comments are not correctly constructed.
	 */
	protected void skipWhiteSpaceAndComments() throws IOException, ESDLReadException {
		// Mark the stream. We do this because we want to check for a non-whitespace
		// character. However, it needs to be unread when it is. We use reset for
		// this. The readahead limit is of course one character.
		reader.mark(1);

		// Check for the end of file.
		try {
			// Go through the characters as long as it is whitespace.
			while("\n\r\t\f ".indexOf(read()) >= 0) {
				// Mark the stream again.
				reader.mark(1);
			}
		}
		catch(ESDLReadException e) {
			// The end of file was reached. This is not whitespace and not comment
			// either, so return.
			return;
		}
		
		// A non-whitespace character was encountered.
		// Reset the stream to unread this character.
		reader.reset();
		
		// Decrease the filepointer.
		position--;
		
		// This could be comment.
		// Whether it is comment:
		boolean iscomment = false;
		
		// Temporarily save the pointer.
		int tempposition = position;
		
		// Check for end of file.
		try {
			// Check for the comment start tag.
			iscomment = checkStartTag("comment");
		}
		catch(ESDLReadException e) {
			// The end of the file was reached. Apparently this was not comment.
			// Reset the stream (it was marked in checkStartTag).
			reader.reset();
			
			// Reset the position pointer.
			position = tempposition;
			
			// No whitespace and no comment. We're done here.
			return;
		}
		
		// Check if the non-whitespace character and co is comment.
		if(iscomment) {
			// It IS comment! Skip it by reading it.
			// Read until we encounter the end tag.
			// If we encounter end-of-file now, then the ESDL is just bad.
			while(!checkEndTag("comment")) {
				// Read the next character and discard it.
				read();
			}
		}
		else {
			// It is not comment. Return.
			return;
		}
		
		// Tail recurse.
		skipWhiteSpaceAndComments();
	}
	
	
	/**
	 * Reads characters from the current position in the reader until as many as the
	 * length of the start tag of the given tag name are read. Then, this string is
	 * compared to the start tag of the given tag name and returned is whether it is
	 * the same. If it is the same, the stream position will be advanced to the point
	 * after the read tag.
	 * 
	 * @param tagname The tag name that the start tag that should be read of.
	 * @return Whether the given tag was read.
	 * @throws IOException When IO errors occur (duh).
	 * @throws ESDLReadException When the end of the stream is reached during
	 * reading.
	 */
	protected boolean checkStartTag(String tagname) throws IOException, ESDLReadException {
		// The read "tag".
		StringBuffer readtag = new StringBuffer();
		
		// Add the enclosing brackets to the tag.
		String tag = "<" + tagname + ">";
		
		// Mark the stream. This is always possible with a BufferedReader.
		// Make the readahead limit equal to the length of tag.
		reader.mark(tag.length());
		
		// Read the "tag" at the current position in the file.
		// So we read as many characters as the tag would be from the stream (which
		// is the length of the tag).
		for(int count = 0; count < tag.length(); count++) {
			// Add the read character to the tag.
			readtag.append(read());
		}
		
		// Compare the given tag with the read tag.
		if(tag.equals(readtag.toString())) {
			// They are the same, so we return true.
			return true;
		}
		else {
			// They are not the same, so we reset the stream and position pointer.
			reader.reset();
			position -= tag.length();

			// Return false to indicate that the tag is not there.
			return false;
		}
	}
	
	
	/**
	 * Reads characters from the current position in the reader until as many as the
	 * length of the end tag of the given tag name are read. Then, this string is
	 * compared to the end tag of the given tag name and returned is whether it is
	 * the same. If it is the same, the stream position will be advanced to the point
	 * after the read tag.
	 * 
	 * @param tagname The tag name that the end tag that should be read of.
	 * @return Whether the given tag was read.
	 * @throws IOException When IO errors occur (duh).
	 * @throws ESDLReadException When the end of the stream is reached during reading.
	 */
	protected boolean checkEndTag(String tagname) throws IOException, ESDLReadException {
		// Add a / in front of the given tag.
		tagname = "/" + tagname;
		
		// Compare the two tags and return whether they are equal. We reuse the
		// readStartTag() code for this :)
		return checkStartTag(tagname);
	}
	
	
	/**
	 * Reads and return the next character in the stream.
	 * 
	 * @return The next character in the stream.
	 * @throws IOException When IO errors occur (duh).
	 * @throws ESDLReadException When an end of file is reached.
	 */
	protected char read() throws IOException, ESDLReadException {
		// Read an int from the stream.
		int readint = reader.read();
		
		// Increase the position in the stream.
		position++;
		
		// Check if the character is "legal" (not whether it represents a number over
		// 18, you pervert!).
		if(readint > 0) {
			// Cast and return the character.
			return (char) readint;
		}
		else {
			// OMG! An illegal character. The only thing that could've happened is an
			// end-of-file (-1 is returned in such a case). So...
			// Decrease the position pointer to the actual end of the stream.
			position--;
			
			// Throw the exception.
			throw getException("Preliminary end of file reached.");
		}
	}
	
	
	/**
	 * Returns an <code>ESDLReadException</code> with the given message, and some
	 * other information, like the place it happened in the file.
	 * 
	 * @param error The error that occured.
	 * @return The <code>ESDLReadException</code> that signifies the given error.
	 */
	protected ESDLReadException getException(String error) {
		// Construct and throw the exception.
		return new ESDLReadException("There is an error in your ESDL description at position " + position + ": " + error);
	}
	
	
	/**
	 * The finalize method. It closes the stream.
	 * 
	 * @see java.lang.Object#finalize()
	 */
	public void finalize() {
		// Catch errors.
		try {
			// Close the stream.
			reader.close();
		}
		catch(IOException e) {
			// Do t3h nothing.
		}
	}
	
	
	/**
	 * Reads the given ESDL file into an ESDL document, and returns that. The data
	 * must have been encoded with UTF-16.
	 * 
	 * @param filename The ESDL file to be read.
	 * @return The <code>SDLDocument</code> that was constructed from the data in the
	 * file.
	 * @throws IOException When IO errors occur.
	 * @throws ESDLReadException When the file is not a correct SDL file.
	 */
	public static SDLDocument readESDLFile(String filename) throws IOException, ESDLReadException {
		// Construct a new ESDLReader around the file specified by filename.
		ESDLReader sdlreader = new ESDLReader(new FileInputStream(filename));
		
		// Read the SDL document and return it.
		return sdlreader.readESDLDocument();  
	}
	
	/**
	 * Reads the given ESDL source into an ESDL document, and returns that. The data
	 * must have been encoded with UTF-16.
	 * 
	 * @param in The <code>InputStream</code> that provides the ESDL to be read.
	 * @return The <code>SDLDocument</code> that was constructed from the data from
	 * the source.
	 * @throws IOException When IO errors occur.
	 * @throws ESDLReadException When the data does not make up a correct ESDL file.
	 */
	public static SDLDocument readESDL(InputStream in) throws IOException, ESDLReadException {
		// Construct a new ESDLReader around the source specified as in.
		ESDLReader sdlreader = new ESDLReader(in);
		
		// Read the SDL document and return it.
		return sdlreader.readESDLDocument();
	}
	
	
	/**
	 * The <code>Reader</code> the ESDL document is read from.
	 */
	protected BufferedReader reader;
	
	/**
	 * The position in the file. This is used to indicate where errors are in an SDL
	 * description.
	 */
	protected int position = 1;
}